#!/usr/bin/env python
#
# correlate.py - The CorrelateAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`.PearsonCorrelateAction` class, which is
an :class:`.Action` that calculates seed-based correlation on 4D
:class:`.Image` overlays.
"""


import threading
import logging

import numpy                  as np
import scipy.spatial.distance as spd

import fsl.data.image               as fslimage
import fsl.utils.idle               as idle
import fsleyes_props                as props
import fsleyes_widgets.utils.status as fslstatus
import fsleyes.strings              as strings
from . import                          base


log = logging.getLogger(__name__)


class CorrelateAction(base.Action):
    """The ``CorrelateAction`` is a base class for the
    :class:`PearsonCorrelateAction` and :class:`PCACorrelateAction` classes,
    which manages adding/removing correlation overlays to/from the
    :class:`.OverlayList`, and manages execution of the correlation.


    When a 4D :class:`.Image` is selected and the ``CorrelateAction`` is
    invoked, a new 3D :class:`.Image` is created and added to the
    :class:`.OverlayList` - this image is referred to as a *correlate overlay*,
    and is used to store and display the correlation values.
    """

    def __init__(self, overlayList, displayCtx, panel):
        """Create a ``CorrelateAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg panel:       The :class:`.CanvasPanel` that owns this action.
        """

        base.Action.__init__(
            self, overlayList, displayCtx, self.__runCorrelateAction)

        self.__panel = panel
        self.__name  = '{}_{}'.format(type(self).__name__, id(self))

        displayCtx .addListener('selectedOverlay',
                                self.__name,
                                self.__selectedOverlayChanged)
        overlayList.addListener('overlays',
                                self.__name,
                                self.__overlayListChanged)

        # TODO Use a single data structure -
        #      using two dicts is fragile
        self.__correlateOverlays = {}
        self.__overlayCorrelates = {}

        # The runCorrelateAction cannot be called
        # more than once at a time - this is used
        # as a semaphore to ensure that this is
        # enforced.
        self.__correlateFlag = threading.Event()

        self.__selectedOverlayChanged()


    def destroy(self):
        """Removes listeners from the :class:`.DisplayContext` and
        :class:`.OverlayList`, and calls :meth:`.Action.destroy`.
        """

        self.displayCtx .removeListener('selectedOverlay', self.__name)
        self.overlayList.removeListener('overlays',        self.__name)
        base.Action.destroy(self)

        self.__correlateOverlays = None
        self.__overlayCorrelates = None


    def __selectedOverlayChanged(self, *a):
        """Called when the selected overlay, or overlay list, changes.

        Enables/disables this action depending on the nature of the selected
        overlay.
        """

        ovl          = self.displayCtx.getSelectedOverlay()
        isCorrOvl    = ovl in self.__overlayCorrelates

        self.enabled = isCorrOvl or  \
                       ((ovl is not None)               and
                        isinstance(ovl, fslimage.Image) and
                        ovl.ndim > 3)


    def __overlayListChanged(self, *a):
        """Called when the :class:`.OverlayList` changes. Makes sure that
        there are no obsolete correlate overlays in the list, and calls
        :meth:`__selectedOverlayChanged`.
        """
        self.__clearCorrelateOverlays()
        self.__selectedOverlayChanged()


    def __clearCorrelateOverlays(self):
        """Called by :meth:`__overlayListChanged`. Clears internal references
        to any obsolete correlate overlays.
        """

        for overlay, corrOvl in list(self.__correlateOverlays.items()):
            if overlay not in self.overlayList or \
               corrOvl not in self.overlayList:
                self.__correlateOverlays.pop(overlay)
                self.__overlayCorrelates.pop(corrOvl)


    def __createCorrelateOverlay(self, overlay, data):
        """Creates a *correlate* overlay for the given ``overlay``, adds
        it to the :class:`.OverlayList`, and initialises some display
        properties.
        """

        display = self.displayCtx.getDisplay(overlay)
        name    = '{}/correlation'.format(display.name)
        corrOvl = fslimage.Image(data, name=name, header=overlay.header)

        self.overlayList.append(corrOvl, overlayType='volume')
        self.__correlateOverlays[overlay] = corrOvl
        self.__overlayCorrelates[corrOvl] = overlay

        corrOpts = self.displayCtx.getOpts(corrOvl)

        with props.suppressAll(corrOpts), \
             props.suppressAll(display):
            corrOpts.cmap              = 'red-yellow'
            corrOpts.negativeCmap      = 'blue-lightblue'
            corrOpts.useNegativeCmap   = True
            corrOpts.displayRange      = [0.05, 1]
            corrOpts.clippingRange.xlo = 0.05

        return corrOvl


    def __runCorrelateAction(self):
        """Called when this :class:`.Action` is invoked. Calculates correlation
        values from the voxel at the current :attr:`.DisplayContext.location`
        (relative to the currently selected overlay) to all other voxels, and
        updates the correlate overlay.

        The correlation calculation and overlay update is performed on a
        separate thread (via :meth:`.idle.run`), with a call to
        :meth:`calculateCorrelation`.
        """

        # Because of the multi-threaded/asynchronous
        # way that this function does its job,
        # allowing it to be called multiple times
        # before prior calls have completed would be
        # very dangerous indeed.
        if self.__correlateFlag.is_set():
            log.debug('Correlate action is already '
                      'running - ignoring request')
            return

        # See if a correlate overlay already exists
        # for the currently selected overlay
        ovl     = self.displayCtx.getSelectedOverlay()
        corrOvl = self.__correlateOverlays.get(ovl, None)

        # If not, check to see if it is a correlate
        # overlay that is selected and, if it is,
        # look up the corresponding source overlay.
        if corrOvl is None:
            if ovl in self.__overlayCorrelates:
                corrOvl = ovl
                ovl     = self.__overlayCorrelates[corrOvl]

        # If corrOvl is still None, it means that
        # there is no correlate overlay for the
        # currently selected overlay. In this case,
        # we'll create a new correlate overlay and
        # add it to the overlay list after the
        # correlation values have been calculated.

        opts = self.displayCtx.getOpts(ovl)
        xyz  = opts.getVoxel(vround=True)

        if xyz is None:
            return

        data = ovl.data[opts.index(atVolume=False)]

        # The correlation calculation is performed
        # on a separate thread. This thread then
        # schedules a function on idle.idle to
        # update the correlation overlay back on the
        # main thread.
        def calcCorr():

            correlations = self.calculateCorrelation(xyz, data)

            # The correlation overlay is updated/
            # created on the main thread.
            def update():

                try:

                    # A correlation overlay already
                    # exists for the source overlay
                    # - update its data
                    if corrOvl is not None:
                        corrOvl[:] = correlations

                    # The correlation overlay hasn't
                    # been created yet - create a
                    # new overlay with the correlation
                    # values.
                    else:
                        self.__createCorrelateOverlay(ovl, correlations)

                finally:
                    fslstatus.clearStatus()
                    self.__correlateFlag.clear()

            idle.idle(update)

        # Protect against more calls
        # while this job is running.
        self.__correlateFlag.set()
        fslstatus.update(strings.messages[self, 'calculating'].format(*xyz))
        idle.run(calcCorr)


    def calculateCorrelation(self, seed, data):
        """Calculates correlation values between the given ``seed`` voxel (an
        ``(x, y, z)`` tuple) and all other voxels. This method must be
        implemented by sub-classes.

        :arg seed: An ``(x, y, z)`` tuple specifying the seed voxel

        :arg data: A 4D ``numpy`` array containing all of the data.

        :returns:  A 3D ``numpy`` array containing the correlation values.
        """
        raise NotImplementedError('calculateCorrelation must be '
                                  'implemented by sub-classes')


class PearsonCorrelateAction(CorrelateAction):
    """The ``PearsonCorrelateAction`` is a :class:`CorrelateAction` which
    calculates Pearson correlation coefficient values between the seed voxel
    and all other voxels.
    """
    def calculateCorrelation(self, seed, data):
        """Calculates Pearson correlation between the data at the specified
        seed voxel, and all other voxels.
        """
        return pearsonCorrelation(seed, data)


def pearsonCorrelation(seed, data):
    """Calculates Pearson correlation between the data at the specified
    seed voxel, and all other voxels.
    """

    x, y, z = seed
    npoints = data.shape[3]

    # the scipy.spatial.distance.cdist
    # function can be used to calculate
    # one-to-many correlation values.
    with np.errstate(invalid='ignore'):
        correlations = 1 - spd.cdist(
            data[x, y, z, :].reshape( 1, npoints),
            data            .reshape(-1, npoints),
            metric='correlation')

    # Set any nans to 0
    correlations[np.isnan(correlations)] = 0

    return correlations.reshape(data.shape[:3])



#!/usr/bin/env python
#
# orthoeditactiontoolbar.py - Ortho edit mode action toolbar.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OrthoEditActionToolbar`, a toolbar used
by the :class:`.OrthoPanel` in edit mode, which contains various buttons
allowing the user to run various edit-related actions.
"""

import wx

import fsleyes_props                 as props

import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.toolbar               as fsltoolbar
import fsleyes.actions               as actions
import fsleyes.icons                 as fslicons
import fsleyes.tooltips              as fsltooltips


class OrthoEditActionToolBar(ctrlpanel.ControlToolBar):
    """The ``OrthoEditActionToolBar`` is a toolbar used by the
    :class:`.OrthoPanel`, which contains buttons allowing the user to:

     - Open the :class:`.OrthoEditSettingsPanel`
     - Create a new :class:`.Image`
     - Undo/redo the last change
     - Clear/fill/erase the current selection
    """

    def __init__(self, parent, overlayList, displayCtx, frame, ortho):
        """Create an ``OrthoEditActionToolBar``.

        :arg parent:      The :mod:`wx` parent object
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        :arg ortho:       The :class:`.OrthoPanel` instance.
        """

        ctrlpanel.ControlToolBar.__init__(self,
                                          parent,
                                          overlayList,
                                          displayCtx,
                                          frame,
                                          height=24,
                                          orient=wx.VERTICAL,
                                          kbFocus=True)

        self.__ortho = ortho

        ortho.addListener('profile', self.name, self.__profileChanged)

        self.__profileChanged()


    def destroy(self):
        """Must be called when this ``OrthoEditAction`` is no longer
        needed. Removes property listeners, and calls the
        :meth:`.ControlToolBar.destroy` method.
        """
        self.__ortho.removeListener('profile', self.name)
        ctrlpanel.ControlToolBar.destroy(self)


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``OrthoEditActionToolBar`` is only intended to be added to
        :class:`.OrthoPanel` views.
        """
        from fsleyes.views.orthopanel import OrthoPanel
        return [OrthoPanel]


    def __profileChanged(self, *a):
        """Called when the :attr:`.ViewPanel.profile` property of the
        :class:`.OrthoPanel` changes. Shows/hides edit controls accordingly.
        """

        self.ClearTools(destroy=True, postevent=False)

        ortho      = self.__ortho
        profile    = ortho.profile
        profileObj = ortho.getCurrentProfile()

        if profile != 'edit':
            return

        tools = []
        nav   = []

        for spec in _TOOLBAR_SPECS:

            if spec == 'div':
                tools.append(fsltoolbar.ToolBarDivider(self,
                                                       height=24,
                                                       orient=wx.HORIZONTAL))
                continue

            if spec.key == 'toggleEditPanel': target = ortho
            else:                             target = profileObj

            widget    = props.buildGUI(self, target, spec)
            navWidget = widget

            if spec.label is not None:
                widget = self.MakeLabelledTool(widget, spec.label)

            tools.append(widget)
            nav  .append(navWidget)

        self.SetTools(tools)
        self.setNavOrder(nav)


_ICONS = {

    'locationFollowsMouse' : [
        fslicons.findImageFile('locationFollowsMouseHighlight24'),
        fslicons.findImageFile('locationFollowsMouse24')],

    'showSelection' : [
        fslicons.findImageFile('showSelectionHighlight24'),
        fslicons.findImageFile('showSelection24')],

    'toggleEditPanel' : [fslicons.findImageFile('editSpannerHighlight24'),
                         fslicons.findImageFile('editSpanner24')],
    'undo'            : fslicons.findImageFile('undo24'),
    'redo'            : fslicons.findImageFile('redo24'),
    'createMask'      : fslicons.findImageFile('new24'),
    'clearSelection'  : fslicons.findImageFile('clearSelection24'),
    'fillSelection'   : fslicons.findImageFile('fillSelection24'),
    'eraseSelection'  : fslicons.findImageFile('eraseSelection24'),
    'copySelection'   : fslicons.findImageFile('copySelection24'),
    'pasteSelection'  : fslicons.findImageFile('pasteSelection24'),
}


_TOOLTIPS = {

    'locationFollowsMouse' : fsltooltips.properties['OrthoEditProfile.'
                                                    'locationFollowsMouse'],
    'showSelection'        : fsltooltips.properties['OrthoEditProfile.'
                                                    'showSelection'],

    'toggleEditPanel' : fsltooltips.actions['OrthoPanel.'
                                            'toggleEditPanel'],
    'undo'            : fsltooltips.actions['OrthoEditProfile.'
                                            'undo'],
    'redo'            : fsltooltips.actions['OrthoEditProfile.'
                                            'redo'],
    'createMask'      : fsltooltips.actions['OrthoEditProfile.'
                                            'createMask'],
    'clearSelection'  : fsltooltips.actions['OrthoEditProfile.'
                                            'clearSelection'],
    'fillSelection'   : fsltooltips.actions['OrthoEditProfile.'
                                            'fillSelection'],
    'eraseSelection'  : fsltooltips.actions['OrthoEditProfile.'
                                            'eraseSelection'],
    'copySelection'   : fsltooltips.actions['OrthoEditProfile.'
                                            'copySelection'],
    'pasteSelection'  : fsltooltips.actions['OrthoEditProfile.'
                                            'pasteSelection'],
}

_TOOLBAR_SPECS = [

    actions.ToggleActionButton(
        'toggleEditPanel',
        actionKwargs={'floatPane' : True},
        icon=_ICONS['toggleEditPanel'],
        tooltip=_TOOLTIPS['toggleEditPanel']),
    actions.ActionButton(
        'createMask',
        icon=_ICONS['createMask'],
        tooltip=_TOOLTIPS['createMask']),

    'div',

    props.Widget(
        'locationFollowsMouse',
        icon=_ICONS['locationFollowsMouse'],
        tooltip=_TOOLTIPS['locationFollowsMouse']),

    'div',


    actions.ActionButton(
        'undo',
        icon=_ICONS['undo'],
        tooltip=_TOOLTIPS['undo']),
    actions.ActionButton(
        'redo',
        icon=_ICONS['redo'],
        tooltip=_TOOLTIPS['redo']),

    'div',

    props.Widget(
        'showSelection',
        icon=_ICONS['showSelection'],
        tooltip=_TOOLTIPS['showSelection'],
        dependencies=['drawMode'],
        enabledWhen=lambda i, m: not m),
    actions.ActionButton(
        'clearSelection',
        icon=_ICONS['clearSelection'],
        tooltip=_TOOLTIPS['clearSelection']),
    actions.ActionButton(
        'fillSelection',
        icon=_ICONS['fillSelection'],
        tooltip=_TOOLTIPS['fillSelection']),
    actions.ActionButton(
        'eraseSelection',
        icon=_ICONS['eraseSelection'],
        tooltip=_TOOLTIPS['eraseSelection']),

    'div',

    actions.ActionButton(
        'copySelection',
        icon=_ICONS['copySelection'],
        tooltip=_TOOLTIPS['copySelection']),
    actions.ActionButton(
        'pasteSelection',
        icon=_ICONS['pasteSelection'],
        tooltip=_TOOLTIPS['pasteSelection']),
]



#!/usr/bin/env python
#
# loadlayout.py - The LoadLayoutAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LoadLayoutAction` class, a FSLeyes action
which allows the user to load a built-in or previously saved layout.
"""


import fsleyes.layouts as layouts
from . import             base


class LoadLayoutAction(base.Action):
    """The ``LoadLayoutAction`` class allows the user to load a specific
    saved layout.
    """


    def __init__(self, overlayList, displayCtx, frame, layout):
        """Create a ``LoadLayoutAction`` instance.

        :arg frame:  The :class:`.FSLeyesFrame`
        :arg layout: Name of the layout to load.
        """

        self.__frame  = frame
        self.__layout = layout
        base.Action.__init__(self, overlayList, displayCtx, self.__loadLayout)


    def __loadLayout(self):
        """Load the layout specified in :meth:`__init__`. """
        layouts.loadLayout(self.__frame, self.__layout)



#!/usr/bin/env python
#
# plotprofile.py - The PlotProfile class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains the :class:`PlotProfile` class, a :class:`.Profile`
for use with :class:`.PlotPanel` views.
"""


from matplotlib.backends.backend_wx import NavigationToolbar2Wx

import fsleyes_props    as props
import fsleyes.profiles as profiles


class PlotProfile(profiles.Profile):
    """The ``PlotProfile`` class is the default interaction profile for
    :class:`.PlotPanel` vies. It provides pan and zoom functionality via
    a single :attr:`.Profile.mode` called ``panzoom``:

     - Left click and drag to pan the plot

     - Right click and drag to zoom the plot.
    """

    def __init__(self,
                 viewPanel,
                 overlayList,
                 displayCtx,
                 extraModes=None):
        """Create a ``PlotProfile``.

        :arg viewPanel:    A :class:`.PlotPanel` instance.

        :arg overlayList:  The :class:`.OverlayList` instance.

        :arg displayCtx:   The :class:`.DisplayContext` instance.

        :arg extraModes:   Extra modes to pass through to the
                           :class:`.Profile` constructor.
        """

        if extraModes is None:
            extraModes = []

        modes = ['panzoom'] + extraModes

        profiles.Profile.__init__(self,
                                  viewPanel,
                                  overlayList,
                                  displayCtx,
                                  modes)

        self.__canvas  = viewPanel.getCanvas()
        self.__axis    = viewPanel.getAxis()

        # Pan/zoom functionality is actually
        # implemented by the NavigationToolbar2Wx
        # class, but the toolbar is not actually
        # shown.
        self.__toolbar = NavigationToolbar2Wx(self.__canvas)
        self.__toolbar.Show(False)

        # This flag keeps track of
        # the toolbar pan state
        self.__panning = False


    def destroy(self):
        """Must be called when this ``PlotProfile`` is no longer needed. Clears
        references and calls the base class ``destroy`` method.
        """

        self.__toolbar.Destroy()

        self.__canvas  = None
        self.__axis    = None
        self.__toolbar = None

        profiles.Profile.destroy(self)


    def getEventTargets(self):
        """Overrides :meth:`.Profile.getEventTargets`. Returns the
        ``matplotlib`` ``Canvas`` object displayed in the :class:`.PlotPanel`.
        """

        return [self.__canvas]


    def __updateAxisLimits(self):
        """Called by the ``panzoom`` ``MouseDrag`` event handlers. Makes sure
        that the :attr:`.PlotPanel.limits` property is up to date.
        """

        xlims = list(self.__axis.get_xlim())
        ylims = list(self.__axis.get_ylim())

        with props.suppress(self.viewPanel, 'limits'):
            self.viewPanel.limits.x = xlims
            self.viewPanel.limits.y = ylims


    def _panzoomModeLeftMouseDown(self, ev, canvas, mousePos, canvasPos):
        """Called on left mouse clicks. Enables panning. """

        if not self.__panning:
            self.__toolbar.pan()
            self.__toolbar.press_pan(self.getMplEvent())
            self.__panning = True


    def _panzoomModeLeftMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Called on left mouse drags. Updates the
        :attr:`.PlotPanel.limits` property - the panning logic is provided
        by the ``matplotlib`` ``NavigationToolbar2wx`` class.
        """
        self.__updateAxisLimits()


    def _panzoomModeLeftMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Called on left mouse up events. Disables panning."""

        if self.__panning:
            self.__toolbar.pan()
            self.__panning = False


    def _panzoomModeRightMouseDown(self, ev, canvas, mousePos, canvasPos):
        """Called on right mouse clicks. Enables zooming. """

        if not self.__panning:
            self.__toolbar.pan()
            self.__toolbar.press_pan(self.getMplEvent())
            self.__panning = True


    def _panzoomModeRightMouseDrag(self, ev, canvas, mousePos, canvasPos):
        """Called on right mouse drags. Updates the
        :attr:`.PlotPanel.limits` property - the zooming logic is provided
        by the ``matplotlib`` ``NavigationToolbar2wx`` class.
        """
        self.__updateAxisLimits()


    def _panzoomModeRightMouseUp(self, ev, canvas, mousePos, canvasPos):
        """Called on right mouse up events. Disables panning. """
        if self.__panning:
            self.__toolbar.pan()
            self.__panning = False



#!/usr/bin/env python
#
# timeseriespanel.py - The TimeSeriesPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`TimeSeriesPanel`, which is a *FSLeyes
view* for displaying time series data from :class:`.Image` overlays.
"""


import logging
import itertools as it

import numpy as np

import wx

import fsl.data.featimage                      as fslfeatimage
import fsl.data.melodicimage                   as fslmelimage
import fsl.data.image                          as fslimage
import fsl.data.mesh                           as fslmesh
import fsleyes_props                           as props

import fsleyes.overlay                         as fsloverlay
import fsleyes.actions                         as actions
import fsleyes.actions.addmaskdataseries       as addmaskdataseries
import fsleyes.strings                         as strings
import fsleyes.plotting.timeseries             as timeseries
import fsleyes.controls.timeseriescontrolpanel as timeseriescontrolpanel
import fsleyes.controls.timeseriestoolbar      as timeseriestoolbar
from . import                                     plotpanel


log = logging.getLogger(__name__)


class TimeSeriesPanel(plotpanel.OverlayPlotPanel):
    """The ``TimeSeriesPanel`` is an :class:`.OverlayPlotPanel` which plots
    time series data from overlays. A ``TimeSeriesPanel`` looks something like
    the following:


    .. image:: images/timeseriespanel.png
       :scale: 50%
       :align: center


    A ``TimeSeriesPanel`` plots one or more :class:`.DataSeries` instances,
    which encapsulate time series data from an overlay. All time series
    related ``DataSeries`` classes are defined in the
    :mod:`.plotting.timeseries` module; these are all sub-classes of the
    :class:`.DataSeries` class - see the :class:`.PlotPanel` and
    :class:`.OverlayPlotPanel` documentation for more details:

    .. autosummary::
       :nosignatures:

       .timeseries.VoxelTimeSeries
       .timeseries.ComplexTimeSeries
       .timeseries.FEATTimeSeries
       .timeseries.MelodicTimeSeries
       .timeseries.MeshTimeSeries


    **Control panels**


    Some *FSLeyes control* panels are associated with the
    :class:`.TimeSeriesPanel`:

    .. autosummary::
       :nosignatures:

       ~fsleyes.controls.plotlistpanel.PlotListPanel
       ~fsleyes.controls.timeseriescontrolpanel.TimeSeriesControlPanel


    The ``TimeSeriesPanel`` defines some :mod:`.actions`, allowing the user
    to show/hide these control panels:

    .. autosummary::
       :nosignatures:

       toggleTimeSeriesToolBar
       toggleTimeSeriesControl


    Some tools are also available, to do various things:

    .. autosummary::
       :nosignatures:

       addMaskDataSeries


    **FEATures**


    The ``TimeSeriesPanel`` has some extra functionality for
    :class:`.FEATImage` overlays. For these overlays, a
    :class:`.FEATTimeSeries` instance is plotted, instead of a regular
    :class:`.TimeSeries` instance. The ``FEATTimeSeries`` class, in turn, has
    the ability to generate more ``TimeSeries`` instances which represent
    various aspects of the FEAT model fit. See the :class:`.FEATTimeSeries`
    and the :class:`.TimeSeriesControlPanel` classes for more details.


    **Melodic features**


    The ``TimeSeriesPanel`` also has some functionality for
    :class:`.MelodicImage` overlays - a :class:`.MelodicTimeSeries` instance
    is used to plot the component time courses for the current component (as
    defined by the :attr:`.NiftiOpts.volume` property).
    """


    usePixdim = props.Boolean(default=False)
    """If ``True``, the X axis data is scaled by the pixdim value of the
    selected overlay (which, for FMRI time series data is typically set
    to the TR time).
    """


    plotMode = props.Choice(('normal', 'demean', 'normalise', 'percentChange'))
    """Options to scale/offset the plotted time courses.

    ================= =======================================================
    ``normal``        The data is plotted with no modifications
    ``demean``        The data is demeaned (i.e. plotted with a mean of 0)
    ``normalise``     The data is normalised to lie in the range ``[-1, 1]``.
    ``percentChange`` The data is scaled to percent changed
    ================= =======================================================
    """


    plotMelodicICs = props.Boolean(default=True)
    """If ``True``, the component time courses are plotted for
    :class:`.MelodicImage` overlays (using a :class:`.MelodicTimeSeries`
    instance). Otherwise, ``MelodicImage`` overlays are treated as regular
    4D :class:`.Image` overlays (a :class:`.VoxelTimeSeries` instance is used).
    """


    def __init__(self, parent, overlayList, displayCtx, frame):
        """Create a ``TimeSeriesPanel``.

        :arg parent:      A :mod:`wx` parent object.
        :arg overlayList: An :class:`.OverlayList` instance.
        :arg displayCtx:  A :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        """

        # If the currently selected image is from
        # a FEAT analysis, and the corresponding
        # filtered_func_data is loaded, enable the
        # initial state of the time course for
        # that filtered_func_data
        featImage = fsloverlay.findFEATImage(
            overlayList,
            displayCtx.getSelectedOverlay())

        if featImage is None: initialState = None
        else:                 initialState = {featImage : True}

        plotpanel.OverlayPlotPanel.__init__(
            self,
            parent,
            overlayList,
            displayCtx,
            frame,
            initialState=initialState)

        self.addListener('plotMode',  self.name, self.draw)
        self.addListener('usePixdim', self.name, self.draw)
        self.addListener('plotMelodicICs',
                         self.name,
                         self.__plotMelodicICsChanged)

        self.__addMaskAction = addmaskdataseries.AddMaskDataSeriesAction(
            overlayList,
            displayCtx,
            self)

        self.addMaskDataSeries.bindProps('enabled', self.__addMaskAction)

        self.initProfile()


    def destroy(self):
        """Removes some listeners, and calls the :meth:`.PlotPanel.destroy`
        method.
        """

        self.removeListener('plotMode',       self.name)
        self.removeListener('usePixdim',      self.name)
        self.removeListener('plotMelodicICs', self.name)

        self.__addMaskAction.destroy()
        self.__addMaskAction = None

        plotpanel.OverlayPlotPanel.destroy(self)


    @actions.toggleControlAction(timeseriescontrolpanel.TimeSeriesControlPanel)
    def toggleTimeSeriesControl(self, floatPane=False):
        """Shows/hides a :class:`.TimeSeriesControlPanel`. See
        :meth:`.ViewPanel.togglePanel`.
        """
        self.togglePanel(timeseriescontrolpanel.TimeSeriesControlPanel,
                         self,
                         location=wx.RIGHT,
                         floatPane=floatPane)


    @actions.toggleControlAction(timeseriestoolbar.TimeSeriesToolBar)
    def toggleTimeSeriesToolBar(self):
        """Shows/hides a :class:`.TimeSeriesToolBar`. See
        :meth:`.ViewPanel.togglePanel`.
        """
        self.togglePanel(timeseriestoolbar.TimeSeriesToolBar, tsPanel=self)


    @actions.action
    def addMaskDataSeries(self):
        """Executes the :class:`AddMaskDataSeriesAction`. """
        self.__addMaskAction()


    def getActions(self):
        """Overrides :meth:`.ActionProvider.getActions`. Returns all of the
        :mod:`.actions` that are defined on this ``TimeSeriesPanel``.
        """
        actionz = [self.screenshot,
                   self.importDataSeries,
                   self.exportDataSeries,
                   None,
                   self.toggleOverlayList,
                   self.togglePlotList,
                   self.toggleTimeSeriesToolBar,
                   self.toggleTimeSeriesControl]

        names = [a.__name__ if a is not None else None for a in actionz]

        return list(zip(names, actionz))


    def getTools(self):
        """Returns a list of tools to be added to the ``FSLeyesFrame`` for
        ``TimeSeriesPanel`` views.
        """
        return [self.addMaskDataSeries]


    def draw(self, *a):
        """Overrides :meth:`.PlotPanel.draw`. Passes some :class:`.DataSeries`
        instances to the :meth:`.PlotPanel.drawDataSeries` method.
        """

        if not self or self.destroyed:
            return

        tss = self.getDataSeriesToPlot()
        for ts in tss:

            # Changing the label might trigger
            # another call to this method, as
            # the PlotPanel might have a listener
            # registered on it. Hence the suppress
            with props.suppress(ts, 'label'):
                ts.label = ts.makeLabel()

        xlabel, ylabel = self.__generateDefaultLabels(tss)

        self.drawDataSeries(extraSeries=tss, xlabel=xlabel, ylabel=ylabel)
        self.drawArtists()


    def createDataSeries(self, overlay):
        """Overrides :meth:`.OverlayPlotPanel.createDataSeries`. Creates and
        returns a :class:`.TimeSeries` instance (or an instance of one of the
        :class:`.TimeSeries` sub-classes) for the specified overlay.

        Returns a tuple containing the following:

          - A :class:`.DataSeries` instance for the given overlay

          - A list of *targets* - objects which have properties that
            influence the state of the ``TimeSeries`` instance.

          - A list of *property names*, one for each target.

        If the given overlay is not compatible (i.e. it has no time series
        data to be plotted), a tuple of ``None`` values is returned.
        """

        displayCtx  = self.displayCtx
        overlayList = self.overlayList
        tsargs      = (overlay, overlayList, displayCtx, self)

        # Is this a mesh?
        if isinstance(overlay, fslmesh.Mesh):
            ts        = timeseries.MeshTimeSeries(*tsargs)
            opts      = displayCtx.getOpts(overlay)
            targets   = [displayCtx, opts]
            propNames = ['location', 'vertexData']

        # Is this a FEAT filtered_func_data image?
        elif isinstance(overlay, fslfeatimage.FEATImage):

            # If the filtered_func for this FEAT analysis
            # has been loaded, we show its time series.
            ts        = timeseries.FEATTimeSeries(*tsargs)
            targets   = [displayCtx]
            propNames = ['location']

        # If this is a melodic IC image, and we are
        # currently configured to plot component ICs,
        # we use a MelodicTimeSeries object.
        elif isinstance(overlay, fslmelimage.MelodicImage) and \
             self.plotMelodicICs:
            ts        = timeseries.MelodicTimeSeries(*tsargs)
            targets   = [displayCtx.getOpts(overlay)]
            propNames = ['volume']

        # Otherwise it's a normal image
        elif isinstance(overlay, fslimage.Image) and overlay.ndim > 3:

            # Is it a complex data image?
            if overlay.iscomplex:
                ts = timeseries.ComplexTimeSeries(*tsargs)

            # Or just a bog-standard 4D image?
            else:
                ts = timeseries.VoxelTimeSeries(*tsargs)

            # listen to volumeDim for
            # images with >4 dimensions
            opts      = displayCtx.getOpts(overlay)
            targets   = [displayCtx, opts]
            propNames = ['location', 'volumeDim']
        else:
            return None, None, None

        ts.colour    = self.getOverlayPlotColour(overlay)
        ts.lineStyle = self.getOverlayPlotStyle(overlay)
        ts.lineWidth = 2
        ts.alpha     = 1

        return ts, targets, propNames


    def prepareDataSeries(self, ts):
        """Overrides :class:`.PlotPanel.prepareDataSeries`. Given a
        :class:`.DataSeries` instance, scales and normalises the x and y data
        according to the current values of the :attr:`usePixdim` and
        :attr:`plotMode` properties.
        """

        xdata, ydata = ts.getData()

        if (xdata is None) or (ydata is None) or (len(xdata) == 0):
            return xdata, ydata

        if self.usePixdim and isinstance(ts.overlay, fslimage.Image):
            if isinstance(ts.overlay, fslmelimage.MelodicImage):
                xdata = xdata * ts.overlay.tr
            else:
                xdata = xdata * ts.overlay.pixdim[3]

            toffset = ts.overlay.nibImage.header.get('toffset', 0)
            xdata  += toffset

        if self.plotMode == 'demean':
            ydata = ydata - ydata.mean()

        elif self.plotMode == 'normalise':
            ymin  = ydata.min()
            ymax  = ydata.max()
            if not np.isclose(ymin, ymax):
                ydata = 2 * (ydata - ymin) / (ymax - ymin) - 1
            else:
                ydata = np.zeros(len(ydata))

        elif self.plotMode == 'percentChange':
            mean  = ydata.mean()
            ydata = 100 * (ydata / mean) - 100

        return xdata, ydata


    def __plotMelodicICsChanged(self, *a):
        """Called when the :attr:`plotMelodicICs` property changes. Re-creates
        the internally cached :class:`.TimeSeries` instances for all
        :class:`.MelodicImage` overlays in the :class:`.OverlayList`.
        """

        for overlay in self.overlayList:
            if isinstance(overlay, fslmelimage.MelodicImage):
                self.clearDataSeries(overlay)

        self.updateDataSeries()
        self.draw()


    def __generateDefaultLabels(self, timeSeries):
        """Called by :meth:`draw`. If the :attr:`.PlotPanel.xlabel` or
        :attr:`.PlotPanel.ylabel` properties are unset, an attempt is made
        to generate default labels.
        """

        xlabel = self.xlabel
        ylabel = self.ylabel

        if xlabel is not None:
            return xlabel, ylabel

        if not self.usePixdim:
            xlabel = strings.nifti['t_unit', -1]
            return xlabel, ylabel

        # If all of the overlays related to the data series being
        # plotted:
        #
        #   - are Images
        #   - have the same time unit (as specified in the nifti header)
        #
        # Then a default label is specified
        #
        # n.b. this is not foolproof, as many
        # non-time 4D images will still set
        # the time units to seconds.
        #
        #
        # TODO Non-Image overlays with associated
        #      time series (e.g. MeshOpts)

        # Get all the unique overlays
        overlays = [ts.overlay for ts in it.chain(timeSeries, self.dataSeries)]
        overlays = set(overlays)

        if not all([isinstance(o, fslimage.Image) for o in overlays]):
            return xlabel, ylabel

        # And all of their time units
        units = [o.timeUnits for o in overlays]

        if len(set(units)) == 1:
            xlabel = strings.nifti.get(('t_unit', units[0]),
                                       'INVALID TIME UNITS')

        return xlabel, ylabel



#!/usr/bin/env python
#
# diagnosticreport.py - The DiagnosticReportAction
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`DiagnosticReportAction`, which allows
the user to generate a report of the current *FSLeyes* state, to send to me
for bug reporting purposes.
"""


import os
import json
import logging
import platform
from collections import OrderedDict

from   fsl.utils.platform import platform as fslplatform
import fsleyes_widgets.utils.status       as status
import fsleyes.strings                    as strings
import fsleyes.state                      as fslstate
from . import                                base


log = logging.getLogger(__name__)


class DiagnosticReportAction(base.Action):
    """The ``DiagnosticReportAction`` generates a JSON-formatted report file
    containing information about the current state of *FSLeyes*. When the this
    :class:`.Action` is run, the user is prompted to select a location to save
    the file Then the report is generated, and written out to the specified
    location.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``DiagnosticReportAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The master :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """

        base.Action.__init__(self, overlayList, displayCtx, self.__action)
        self.__frame = frame


    def __action(self):
        """This method is the guts of the ``DiagnosticReportAction``. It does
        the following:

          1. Prompts the user to select a location to save the report file.

          2. Generates the report.

          3. Formats the report as JSON.

          4. Saves the report to the specified location.
        """

        import wx

        dlg = wx.FileDialog(
            self.__frame,
            message=strings.titles[self, 'saveReport'],
            defaultDir=os.getcwd(),
            defaultFile='report.txt',
            style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)

        if dlg.ShowModal()  != wx.ID_OK:
            return

        path = dlg.GetPath()

        status.update('Writing diagnostic report to {}...'.format(path))

        report = self.__generateReport()
        report = self.__formatReport(report)

        log.debug('Diagnostic report:\n{}'.format(report))

        with open(path, 'wt') as f:
            f.write(report)


    def __generateReport(self):
        """Generates and returns a *report*, a hierarchical dictionary
        containing information about the current system and *FSLeyes*
        state.
        """

        import fsleyes.version as version

        report   = OrderedDict()
        overlays = []

        for i, ovl in enumerate(self.overlayList):
            overlays.append(OrderedDict([
                ('type',   type(ovl).__name__),
                ('name',   ovl.name),
                ('source', ovl.dataSource)]))

        report['Platform']    = platform.platform()
        report['Python']      = '{}  {}'.format(platform.python_version(),
                                                platform.python_compiler())
        report['Version']     = version.__version__
        report['OpenGL']      = self.__openGLReport()
        report['Settings']    = self.__settingsReport()
        report['State']       = fslstate.getState(self.__frame)

        return report


    def __settingsReport(self):
        """Creates and returns a dictionary containing:

         - *FSLeyes* settings stored via the :mod:`.settings` module
        """

        import fsl.utils.settings as fslsettings

        report   = OrderedDict()
        settings = fslsettings.readAll()

        for k, v in settings.items():
            report[k] = v

        return report


    def __openGLReport(self):
        """Creates and returns a dictionary containing information about the
        OpenGL platform.
        """

        import OpenGL.GL as gl

        report = OrderedDict()

        version    = gl.glGetString( gl.GL_VERSION)   .decode('ascii')
        renderer   = gl.glGetString( gl.GL_RENDERER)  .decode('ascii')
        extensions = gl.glGetString( gl.GL_EXTENSIONS).decode('ascii')

        report['Version']       = version
        report['Compatibility'] = fslplatform.glVersion
        report['Renderer']      = renderer
        report['Texture size']  = str(gl.glGetInteger(gl.GL_MAX_TEXTURE_SIZE))
        report['Extensions']    = extensions.split(' ')

        return report


    def __formatReport(self, reportDict):
        """Converts the given hierarchical dictionary to a JSON-formatted
        string.
        """
        return json.dumps(reportDict, indent=2)



#!/usr/bin/env python
#
# perspectives.py - Deprecated - use layouts.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""Deprecated - see the :mod:`.layouts` module. """


import fsl.utils.deprecated as deprecated

from . import layouts


def deprecate(func):
    return deprecated.deprecated(
        '0.24.0', '1.0.0', 'Use the fsleyes.layout module')(func)


getAllPerspectives     = deprecate(layouts.getAllLayouts)
loadPerspective        = deprecate(layouts.loadLayout)
applyPerspective       = deprecate(layouts.applyLayout)
savePerspective        = deprecate(layouts.saveLayout)
removePerspective      = deprecate(layouts.removeLayout)
serialisePerspective   = deprecate(layouts.serialiseLayout)
deserialisePerspective = deprecate(layouts.deserialiseLayout)
BUILT_IN_PERSPECTIVES  =           layouts.BUILT_IN_LAYOUTS



#!/usr/bin/env python
#
# plotlistpanel.py - The PlotListPanel class
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`PlotListPanel` a *FSLeyes control* panel
which allows the user to add/remove :class:`.DataSeries` from an
:class:`.OverlayPlotPanel`.
"""


import numpy as np

import wx

import fsleyes_props                 as props
import fsleyes_widgets.elistbox      as elistbox

import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.tooltips              as fsltooltips
import fsleyes.strings               as strings


class PlotListPanel(ctrlpanel.ControlPanel):
    """The ``PlotListPanel`` is a *FSLeyes control* panel for use with
    :class:`.OverlayPlotPanel` views. It allows the user to add and remove
    :class:`.DataSeries` instances from the :attr:`.PlotPanel.dataSeries`
    list.


    For every :class:`.DataSeries` instance in the
    :attr:`.PlotPanel.dataSeries` list of the :class:`.OverlayPlotPanel`, the
    ``PlotListPanel`` creates a :class:`.DataSeriesWidget`, which allows the
    user to change the display settings of the :class:`.DataSeries`
    instance. A :class:`.EditableListBox` is used to display the labels for
    each :class:`.DataSeries` instance, and the associated
    :class:`.DataSeriesWidget` controls.
    """


    def __init__(self, parent, overlayList, displayCtx, frame, plotPanel):
        """Create a ``PlotListPanel``.

        :arg parent:      The :mod:`wx` parent object.

        :arg overlayList: The :class:`.OverlayList`.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg frame:       The :class:`.FSLeyesFrame` instance.

        :arg plotPanel:   The :class:`.OverlayPlotPanel` associated with this
                          ``PlotListPanel``.
        """

        ctrlpanel.ControlPanel.__init__(
            self, parent, overlayList, displayCtx, frame)

        self.__plotPanel = plotPanel
        self.__dsList    = elistbox.EditableListBox(
            self, style=(elistbox.ELB_NO_MOVE  |
                         elistbox.ELB_EDITABLE |
                         elistbox.ELB_WIDGET_RIGHT))

        self.__sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.__sizer)

        self.__sizer.Add(self.__dsList, flag=wx.EXPAND, proportion=1)

        self.__dsList.Bind(elistbox.EVT_ELB_ADD_EVENT,    self.__onListAdd)
        self.__dsList.Bind(elistbox.EVT_ELB_REMOVE_EVENT, self.__onListRemove)
        self.__dsList.Bind(elistbox.EVT_ELB_EDIT_EVENT,   self.__onListEdit)
        self.__dsList.Bind(elistbox.EVT_ELB_SELECT_EVENT, self.__onListSelect)

        self.__plotPanel.addListener('dataSeries',
                                     self.name,
                                     self.__dataSeriesChanged)

        self.__dataSeriesChanged()
        self.Layout()
        self.SetMinSize(self.__sizer.GetMinSize())


    def destroy(self):
        """Must be called when this ``PlotListPanel`` is no longer
        needed. Removes some property listeners, and calls the
        :meth:`.ControlPanel.destroy` method.
        """

        self.__plotPanel.removeListener('dataSeries', self.name)
        self.__plotPanel = None
        self.__dsList.Clear()
        ctrlpanel.ControlPanel.destroy(self)


    def __dataSeriesChanged(self, *a):
        """Called when the :attr:`.PlotPanel.dataSeries` list of the
        :class:`.OverlayPlotPanel` changes. Updates the list of
        :class:`.TimeSeriesWidget` controls.
        """

        self.__dsList.Clear()

        for ds in self.__plotPanel.dataSeries:
            widg = DataSeriesWidget(self, ds)
            self.__dsList.Append(
                ds.label,
                clientData=ds,
                tooltip=fsltooltips.properties[ds, 'label'],
                extraWidget=widg)


    def __onListAdd(self, ev):
        """Called when the user pushes the *add* button on the
        :class:`.EditableListBox`.  Adds the :class:`.DataSeries` associated
        with the currently selected overlay to the
        :attr:`.PlotPanel.dataSeries` list of the :class:`.OverlayPlotPanel`.
        """

        self.__plotPanel.addDataSeries()


    def __onListEdit(self, ev):
        """Called when the user edits a label on the
        :class:`.EditableListBox`. Updates the :attr:`.DataSeries.label`
        property of the corresponding :class:`DataSeries` instance.
        """
        ev.data.label = ev.label


    def __onListSelect(self, ev):
        """Called when the user selects an item in the
        :class:`.EditableListBox`. Sets the
        :attr:`.DisplayContext.selectedOverlay` to the overlay associated
        with the corresponding :class:`.DataSeries` instance.
        """

        ds      = ev.data
        overlay = ds.overlay

        if overlay is None:
            return

        opts = self.displayCtx.getOpts(overlay)

        self.displayCtx.selectedOverlay = self.overlayList.index(overlay)

        # See hacky things in __onListAdd
        if hasattr(ds, '_volume'):
            opts.volume = ds._volume

        elif hasattr(ds, '_location'):
            voxLoc = np.array(ds._location)
            disLoc = opts.transformCoords([voxLoc], 'voxel', 'display')[0]
            self.displayCtx.location = disLoc


    def __onListRemove(self, ev):
        """Called when the user removes an item from the
        :class:`.EditableListBox`. Removes the corresponding
        :class:`.DataSeries` instance from the :attr:`.PlotPanel.dataSeries`
        list of the :class:`.OverlayPlotPanel`.
        """
        with props.skip(self.__plotPanel, 'dataSeries', self.name):
            self.__plotPanel.dataSeries.remove(ev.data)


class DataSeriesWidget(wx.Panel):
    """The ``DataSeriesWidget`` class is a panel which contains controls
    that modify the properties of a :class:`.DataSeries` instance. A
    ``DataSeriesWidget`` is created by the :class:`PlotListPanel` for
    every ``DataSeries`` in the :attr:`.PlotPanel.dataSeries` list.
    """


    def __init__(self, parent, dataSeries):
        """Create a ``DataSeriesWidget``.

        :arg parent:     The :mod:`wx` parent object.

        :arg dataSeries: The :class:`.DataSeries` instance.
        """

        wx.Panel.__init__(self, parent)

        self.__enabled   = props.makeWidget(self, dataSeries, 'enabled')
        self.__colour    = props.makeWidget(self, dataSeries, 'colour')
        self.__lineWidth = props.makeWidget(self, dataSeries, 'lineWidth')
        self.__lineStyle = props.makeWidget(
            self,
            dataSeries,
            'lineStyle',
            labels=strings.choices['DataSeries.lineStyle'])

        self.__enabled.SetToolTip(
            wx.ToolTip(fsltooltips.properties[dataSeries, 'enabled']))
        self.__colour.SetToolTip(
            wx.ToolTip(fsltooltips.properties[dataSeries, 'colour']))
        self.__lineWidth.SetToolTip(
            wx.ToolTip(fsltooltips.properties[dataSeries, 'lineWidth']))
        self.__lineStyle.SetToolTip(
            wx.ToolTip(fsltooltips.properties[dataSeries, 'lineStyle']))

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(self.__sizer)

        self.__sizer.Add(self.__enabled)
        self.__sizer.Add(self.__colour)
        self.__sizer.Add(self.__lineWidth)
        self.__sizer.Add(self.__lineStyle)

        self.Layout()



#!/usr/bin/env python
#
# loadperspective.py - deprecated.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""Deprecated - see :mod:`.loadlayout`. """


import fsl.utils.deprecated as deprecated


from . import loadlayout


class LoadPerspectiveAction(loadlayout.LoadLayoutAction):

    @deprecated.deprecated('0.24.0', '1.0.0', 'use LoadLayoutAction')
    def __init__(self, *args, **kwargs):
        loadlayout.LoadLayoutAction.__init__(self, *args, **kwargs)



#!/usr/bin/env python
#
# addroihistogram.py - The AddROIHistogramAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`AddROIHistogramAction` class, an
action used by the :class:`.HistogramPanel`.
"""


import          wx
import numpy as np

import fsl.data.image                   as fslimage

import fsleyes.strings                  as strings
import fsleyes.plotting.dataseries      as dataseries
import fsleyes.plotting.histogramseries as histogramseries

from . import                         base
from . import                         addmaskdataseries


class AddROIHistogramAction(base.Action):
    """The ``AddROIHistogramAction`` class is used by the
    :class:`.HistogramPanel`.

    It performs a very similar task to the :class:`.AddMaskDataSeriesAction` -
    the user selects a binary mask, the data within the base image is extracted
    for that mask, and the histogram of that data is added to the plot.
    """


    def __init__(self, overlayList, displayCtx, plotPanel):
        """Create an ``AddROIHistogramAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg plotPanel:   The :class:`.HistogramPanel`.
        """

        base.Action.__init__(
            self, overlayList, displayCtx, self.__addROIHistogram)

        self.__plotPanel  = plotPanel
        self.__name       = '{}_{}'.format(type(self).__name__, id(self))
        self.__roiOptions = []

        overlayList.addListener('overlays',
                                self.__name,
                                self.__overlayListChanged)
        displayCtx .addListener('selectedOverlay',
                                self.__name,
                                self.__overlayListChanged)

        self.__overlayListChanged()


    def destroy(self):
        """Must be called when this ``AddROIHistogramAction`` is no
        longer in use.
        """
        self.overlayList.removeListener('overlays',        self.__name)
        self.displayCtx .removeListener('selectedOverlay', self.__name)
        self.__plotPanel  = None
        self.__roiOptions = None
        base.Action.destroy(self)


    def __overlayListChanged(self, *a):
        """Called when the :class:`.OverlayList` or the
        :attr:`.DisplayContext.selectedOverlay` changes. Updates a list of
        valid mask images for the currently selected overlay.
        """

        overlay = self.displayCtx.getSelectedOverlay()

        if (len(self.overlayList) == 0 or
           (not isinstance(overlay, fslimage.Image))):
            self.enabled = False
            return

        self.__roiOptions = [o for o in self.overlayList if
                             isinstance(o, fslimage.Image) and
                             o is not overlay              and
                             o.sameSpace(overlay)]

        self.enabled = len(self.__roiOptions)  > 0



    def __addROIHistogram(self):
        """Prompts the user to select an ROI mask, calculates the histogram
        of that mask on the currently selected overlay, and adds the result
        to the ``HistogramPanel``.
        """

        overlay    = self.displayCtx.getSelectedOverlay()
        opts       = self.displayCtx.getOpts(overlay)
        roiOptions = self.__roiOptions

        frame   = wx.GetApp().GetTopWindow()
        msg     = strings.messages[self, 'selectMask'].format(overlay.name)
        title   = strings.titles[  self, 'selectMask'].format(overlay.name)

        dlg = addmaskdataseries.MaskDialog(
            frame,
            [o.name for o in roiOptions],
            title=title,
            message=msg,
            checkbox=False)

        if dlg.ShowModal() != wx.ID_OK:
            return

        maskimg = roiOptions[dlg.GetChoice()]
        mask    = maskimg[:] > 0

        if overlay.ndim > 3: data = overlay[opts.index()][mask]
        else:                data = overlay[mask]

        count           = self.__plotPanel.histType == 'count'
        drange          = (np.nanmin(data), np.nanmax(data))
        nbins           = histogramseries.autoBin(data, drange)
        xdata, ydata, _ = histogramseries.histogram(data,
                                                    nbins,
                                                    drange,
                                                    drange,
                                                    includeOutliers=False,
                                                    count=count)

        ds           = dataseries.DataSeries(overlay,
                                             self.overlayList,
                                             self.displayCtx,
                                             self.__plotPanel)
        ds.colour    = self.__plotPanel.getOverlayPlotColour(overlay)
        ds.lineStyle = self.__plotPanel.getOverlayPlotStyle(overlay)
        ds.lineWidth = 2
        ds.alpha     = 1
        ds.label     = '{} [mask: {}]'.format(overlay.name, maskimg.name)

        # We have to run the data through
        # prepareDataSeries to preprocess
        # (e.g. smooth) it
        ds.setData(xdata, ydata)
        ds.setData(*self.__plotPanel.prepareDataSeries(ds))

        self.__plotPanel.dataSeries.append(ds)



#!/usr/bin/env python
#
# annotations.py - 2D annotations on a SliceCanvas.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`Annotations` class, which implements
functionality to draw 2D OpenGL annotations on a :class:`.SliceCanvas`.


The :class:`Annotations` class is used by the :class:`.SliceCanvas` and
:class:`.LightBoxCanvas` classes, and users of those class, to annotate the
canvas.


All annotations derive from the :class:`AnnotationObject` base class. The
following annotation types are defined:

.. autosummary::
   :nosignatures:

   Line
   Rect
   VoxelGrid
   VoxelSelection
"""


import logging
import time

import numpy       as np
import OpenGL.GL   as gl

import fsleyes.gl.globject       as globject
import fsleyes.gl.routines       as glroutines
import fsleyes.gl.resources      as glresources
import fsleyes.gl.textures       as textures
import fsleyes.gl.textures.data  as texdata
import fsl.transform.affine      as affine


log = logging.getLogger(__name__)


class Annotations(object):
    """An :class:`Annotations` object provides functionality to draw 2D
    annotations on a :class:`.SliceCanvas`. Annotations may be enqueued via
    any of the :meth:`line`, :meth:`rect`, :meth:`grid`, :meth:`selection` or
    :meth:`obj`, methods.


    A call to :meth:`draw` will then draw each of the queued annotations on
    the canvas, and clear the queue.


    If an annotation is to be persisted, it can be enqueued, as above, but
    passing ``hold=True`` to the queueing method.  The annotation will then
    remain in the queue until it is removed via :meth:`dequeue`, or the
    entire annotations queue is cleared via :meth:`clear`.


    Annotations can be queued by one of the helper methods on the
    :class:`Annotations` object (e.g. :meth:`line` or :meth:`rect`), or by
    manually creating an :class:`AnnotationObject` and passing it to the
    :meth:`obj` method.
    """


    def __init__(self, canvas, xax, yax):
        """Creates an :class:`Annotations` object.

        :arg canvas: The :class:`.SliceCanvas` that owns this
                     ``Annotations`` object.

        :arg xax:    Index of the display coordinate system axis that
                     corresponds to the horizontal screen axis.

        :arg yax:    Index of the display coordinate system axis that
                     corresponds to the vertical screen axis.
        """

        self.__q      = []
        self.__holdq  = []
        self.__xax    = xax
        self.__yax    = yax
        self.__zax    = 3 - xax - yax
        self.__canvas = canvas


    @property
    def canvas(self):
        """Returns a ref to the canvas that owns this ``Annotations`` instance.
        """
        return self.__canvas


    def setAxes(self, xax, yax):
        """This method must be called if the display orientation changes.  See
        :meth:`__init__`.
        """

        self.__xax = xax
        self.__yax = yax
        self.__zax = 3 - xax - yax


    def getDisplayBounds(self):
        """Returns a tuple containing the ``(xmin, xmax, ymin, ymax)`` display
        bounds of the ``SliceCanvas`` that owns this ``Annotations`` object.
        """
        return self.__canvas.opts.displayBounds


    def line(self, *args, **kwargs):
        """Queues a line for drawing - see the :class:`Line` class. """
        hold = kwargs.pop('hold', False)
        obj  = Line(self, *args, **kwargs)

        return self.obj(obj, hold)


    def rect(self, *args, **kwargs):
        """Queues a rectangle for drawing - see the :class:`Rectangle` class.
        """
        hold = kwargs.pop('hold', False)
        obj  = Rect(self, *args, **kwargs)

        return self.obj(obj, hold)


    def grid(self, *args, **kwargs):
        """Queues a voxel grid for drawing - see the :class:`VoxelGrid` class.
        """
        hold = kwargs.pop('hold', False)
        obj  = VoxelGrid(self, *args, **kwargs)

        return self.obj(obj, hold)


    def selection(self, *args, **kwargs):
        """Queues a selection for drawing - see the :class:`VoxelSelection`
        class.
        """
        hold = kwargs.pop('hold', False)
        obj  = VoxelSelection(self, *args, **kwargs)

        return self.obj(obj, hold)


    def text(self, *args, **kwargs):
        """Queues a text annotation for drawing - see the :class:`Text`
        class.
        """
        hold = kwargs.pop('hold', False)
        obj  = Text(self, *args, **kwargs)

        return self.obj(obj, hold)


    def obj(self, obj, hold=False):
        """Queues the given :class:`AnnotationObject` for drawing.

        :arg hold: If ``True``, the given ``AnnotationObject`` will be kept in
                   the queue until it is explicitly removed. Otherwise (the
                   default), the object will be removed from the queue after
                   it has been drawn.
        """

        if hold: self.__holdq.append(obj)
        else:    self.__q    .append(obj)

        return obj


    def dequeue(self, obj, hold=False):
        """Removes the given :class:`AnnotationObject` from the queue, but
        does not call its :meth:`.GLObject.destroy` method - this is the
        responsibility of the caller.
        """

        if hold:
            try:               self.__holdq.remove(obj)
            except ValueError: pass
        else:
            try:               self.__q.remove(obj)
            except ValueError: pass


    def clear(self):
        """Clears both the normal queue and the persistent (a.k.a. ``hold``)
        queue, and calls the :meth:`.GLObject.destroy` method on every object
        in the queue.
        """

        for obj in self.__q:     obj.destroy()
        for obj in self.__holdq: obj.destroy()

        self.__q     = []
        self.__holdq = []


    def draw(self, zpos, xform=None, skipHold=False):
        """Draws all enqueued annotations.

        :arg zpos:     Position along the Z axis, above which all annotations
                       should be drawn.

        :arg xform:    Transformation matrix which should be applied to all
                       objects.

        :arg skipHold: Do not draw items on the hold queue - only draw one-off
                       items.
        """

        if not skipHold: objs = self.__holdq + self.__q
        else:            objs = self.__q

        if xform is not None:
            gl.glMatrixMode(gl.GL_MODELVIEW)
            gl.glPushMatrix()
            gl.glMultMatrixf(xform.ravel('F'))

        drawTime = time.time()
        axes     = (self.__xax, self.__yax, self.__zax)

        for obj in objs:

            if obj.expired(drawTime):                    continue
            if not obj.enabled:                          continue
            if obj.zmin is not None and zpos < obj.zmin: continue
            if obj.zmax is not None and zpos > obj.zmax: continue

            if obj.xform is not None:
                gl.glMatrixMode(gl.GL_MODELVIEW)
                gl.glPushMatrix()
                gl.glMultMatrixf(obj.xform.ravel('F'))

            if obj.colour is not None:

                if len(obj.colour) == 3: colour = list(obj.colour) + [1.0]
                else:                    colour = list(obj.colour)

                gl.glColor4f(*colour)

            if obj.width is not None:
                gl.glLineWidth(obj.width)

            try:
                obj.preDraw()
                obj.draw2D(zpos, axes)
                obj.postDraw()
            except Exception as e:
                log.warn('{}'.format(e), exc_info=True)

            if obj.xform is not None:
                gl.glPopMatrix()

        if xform is not None:
            gl.glMatrixMode(gl.GL_MODELVIEW)
            gl.glPopMatrix()

        # Clear the regular queue after each draw
        self.__q = []


class AnnotationObject(globject.GLSimpleObject):
    """Base class for all annotation objects. An ``AnnotationObject`` is drawn
    by an :class:`Annotations` instance. The ``AnnotationObject`` contains some
    attributes which are common to all annotation types:

    ============ =============================================================
    ``colour``   Annotation colour
    ``width``    Annotation line width (if the annotation is made up of lines)
    ``xform``    Custom transformation matrix to apply to annotation vertices.
    ``expiry``   Time (in seconds) after which the annotation will expire and
                 not be drawn.
    ``zmin``     Minimum z value below which this annotation will not be
                 drawn.
    ``zmax``     Maximum z value above which this annotation will not be
                 drawn.
    ``creation`` Time of creation.
    ============ =============================================================

    All of these attributes can be modified directly, after which you should
    trigger a draw on the owning ``SliceCanvas`` to refresh the annotation.
    You shouldn't touch the ``expiry`` or ``creation`` attributes though.

    Subclasses must, at the very least, override the
    :meth:`globject.GLObject.draw2D` method.
    """

    def __init__(self,
                 annot,
                 xform=None,
                 colour=None,
                 width=None,
                 enabled=True,
                 expiry=None,
                 zmin=None,
                 zmax=None):
        """Create an ``AnnotationObject``.

        :arg annot:   The :class:`Annotations` object that created this
                      ``AnnotationObject``.

        :arg xform:   Transformation matrix which will be applied to all
                      vertex coordinates.

        :arg colour:  RGB/RGBA tuple specifying the annotation colour.

        :arg width:   Line width to use for the annotation.

        :arg enabled: Initially enabled or disabled.

        :arg expiry:  Time (in seconds) after which this annotation should be
                      expired and not drawn.

        :arg zmin:    Minimum z value below which this annotation should not
                      be drawn.
        :arg zmax:    Maximum z value above which this annotation should not
                      be drawn.
        """
        globject.GLSimpleObject.__init__(self, False)

        self.annot    = annot
        self.colour   = colour
        self.enabled  = enabled
        self.width    = width
        self.xform    = xform
        self.expiry   = expiry
        self.zmin     = zmin
        self.zmax     = zmax
        self.creation = time.time()

        if self.xform is not None:
            self.xform = np.array(self.xform, dtype=np.float32)


    def resetExpiry(self):
        """Resets the expiry for this ``AnnotationObject`` so that it is
        valid from the current time.
        """
        self.creation = time.time()


    def expired(self, now):
        """Returns ``True`` if this ``Annotation`` has expired, ``False``
        otherwise.

        :arg now: The current time
        """
        if self.expiry is None:
            return False

        return (self.creation + self.expiry) < now


    def preDraw(self, *args, **kwargs):
        gl.glEnableClientState(gl.GL_VERTEX_ARRAY)


    def postDraw(self, *args, **kwargs):
        gl.glDisableClientState(gl.GL_VERTEX_ARRAY)


class Line(AnnotationObject):
    """The ``Line`` class is an :class:`AnnotationObject` which represents a
    2D line.
    """

    def __init__(self, annot, xy1, xy2, *args, **kwargs):
        """Create a ``Line`` annotation.

        The ``xy1`` and ``xy2`` coordinate tuples should be in relation to the
        axes which map to the horizontal/vertical screen axes on the target
        canvas.

        :arg annot: The :class:`Annotations` object that owns this ``Line``.

        :arg xy1:   Tuple containing the (x, y) coordinates of one endpoint.

        :arg xy2:   Tuple containing the (x, y) coordinates of the second
                    endpoint.

        All other arguments are passed through to
        :meth:`AnnotationObject.__init__`.
        """
        AnnotationObject.__init__(self, annot, *args, **kwargs)
        self.xy1 = xy1
        self.xy2 = xy2


    def draw2D(self, zpos, axes):
        """Draws this ``Line`` annotation. """

        xax, yax, zax = axes

        idxs                 = np.arange(2,     dtype=np.uint32)
        verts                = np.zeros((2, 3), dtype=np.float32)
        verts[0, [xax, yax]] = self.xy1
        verts[1, [xax, yax]] = self.xy2
        verts[:, zax]        = zpos
        verts                = verts.ravel('C')

        gl.glVertexPointer(3, gl.GL_FLOAT, 0, verts)
        gl.glDrawElements(gl.GL_LINES, len(idxs), gl.GL_UNSIGNED_INT, idxs)


class Rect(AnnotationObject):
    """The ``Rect`` class is an :class:`AnnotationObject` which represents a
    2D rectangle.
    """

    def __init__(self,
                 annot,
                 xy,
                 w,
                 h,
                 filled=False,
                 fillColour=None,
                 *args,
                 **kwargs):
        """Create a :class:`Rect` annotation.

        :arg annot:      The :class:`Annotations` object that owns this
                         ``Rect``.

        :arg xy:         Tuple specifying bottom left of the rectangle, in
                         the display coordinate system.

        :arg w:          Rectangle width.

        :arg h:          Rectangle height.

        :arg filled:     If ``True``, the rectangle is filled with the
                         ``fillColour``.

        :arg fillColour: If ``filled=True``, the colour to fill the rectangle
                         with. Defaults to a transparent version of the
                         ``colour``.

        All other arguments are passed through to
        :meth:`AnnotationObject.__init__`.
        """
        AnnotationObject.__init__(self, annot, *args, **kwargs)

        self.xy         = xy
        self.w          = w
        self.h          = h
        self.filled     = filled
        self.fillColour = fillColour


    def draw2D(self, zpos, axes):
        """Draws this ``Rectangle`` annotation. """

        if self.w == 0 or self.h == 0:
            return

        xax, yax, zax = axes
        xy            = self.xy
        w             = self.w
        h             = self.h

        bl = [xy[0],     xy[1]]
        br = [xy[0] + w, xy[1]]
        tl = [xy[0],     xy[1] + h]
        tr = [xy[0] + w, xy[1] + h]

        self.__drawRect(zpos, xax, yax, zax, bl, br, tl, tr)

        if self.filled:
            self.__drawFill(zpos, xax, yax, zax, bl, br, tl, tr)


    def __drawFill(self, zpos, xax, yax, zax, bl, br, tl, tr):
        """Draw a filled version of the rectangle. """

        fillColour = self.fillColour

        if fillColour is None:
            if self.colour is not None:
                fillColour = list(self.colour[:3])
            else:
                fillColour = [1, 1, 1]

        if len(fillColour) == 3:
            fillColour = list(fillColour) + [0.2]

        idxs  = np.array([0, 1, 2, 2, 1, 3], dtype=np.uint32)
        verts = np.zeros((4, 3),             dtype=np.float32)

        verts[0, [xax, yax]] = bl
        verts[1, [xax, yax]] = br
        verts[2, [xax, yax]] = tl
        verts[3, [xax, yax]] = tr
        verts[:,  zax]       = zpos
        verts                = verts.ravel('C')

        # I'm assuming that glPolygonMode
        # is already set to GL_FILL
        gl.glColor4f(*fillColour)
        gl.glVertexPointer(3, gl.GL_FLOAT, 0, verts)
        gl.glDrawElements(gl.GL_TRIANGLES, len(idxs), gl.GL_UNSIGNED_INT, idxs)


    def __drawRect(self, zpos, xax, yax, zax, bl, br, tl, tr):
        """Draw the rectangle outline. """

        idxs  = np.array([0, 1, 2, 3, 0, 2, 1, 3], dtype=np.uint32)
        verts = np.zeros((4, 3),                   dtype=np.float32)

        verts[0, [xax, yax]] = bl
        verts[1, [xax, yax]] = br
        verts[2, [xax, yax]] = tl
        verts[3, [xax, yax]] = tr
        verts[:,  zax]       = zpos
        verts                = verts.ravel('C')

        gl.glVertexPointer(3, gl.GL_FLOAT, 0, verts)
        gl.glDrawElements(gl.GL_LINES, len(idxs), gl.GL_UNSIGNED_INT, idxs)


class VoxelGrid(AnnotationObject):
    """The ``VoxelGrid`` is an :class:`AnnotationObject` which represents a
    collection of selected voxels. See also the :class:`VoxelSelection`
    annotation.

    Each selected voxel is highlighted with a rectangle around its border.
    """


    def __init__(self,
                 annot,
                 selectMask,
                 displayToVoxMat,
                 voxToDisplayMat,
                 offsets=None,
                 *args,
                 **kwargs):
        """Create a ``VoxelGrid`` annotation.

        :arg annot:           The :class:`Annotations` object that owns this
                              ``VoxelGrid``.

        :arg selectMask:      A 3D numpy array, the same shape as the image
                              being annotated (or a sub-space of the image -
                              see the ``offsets`` argument),  which is
                              interpreted as a mask array - values which are
                              ``True`` denote selected voxels.

        :arg displayToVoxMat: A transformation matrix which transforms from
                              display space coordinates into voxel space
                              coordinates.

        :arg voxToDisplayMat: A transformation matrix which transforms from
                              voxel coordinates into display space
                              coordinates.

        :arg offsets:         If ``None`` (the default), the ``selectMask``
                              must have the same shape as the image data
                              being annotated. Alternately, you may set
                              ``offsets`` to a sequence of three values,
                              which are used as offsets for the xyz voxel
                              values. This is to allow for a sub-space of
                              the full image space to be annotated.
        """

        kwargs['xform'] = voxToDisplayMat
        AnnotationObject.__init__(self, annot, *args, **kwargs)

        if offsets is None:
            offsets = [0, 0, 0]

        self.displayToVoxMat = displayToVoxMat
        self.selectMask      = selectMask
        self.offsets         = offsets


    def draw2D(self, zpos, axes):
        """Draws this ``VoxelGrid`` annotation. """

        xax, yax, zax = axes
        dispLoc       = [0] * 3
        dispLoc[zax]  = zpos
        voxLoc        = affine.transform([dispLoc], self.displayToVoxMat)[0]

        vox = int(round(voxLoc[zax]))

        restrictions = [slice(None)] * 3
        restrictions[zax] = slice(vox - self.offsets[zax],
                                  vox - self.offsets[zax] + 1)

        xs, ys, zs = np.where(self.selectMask[restrictions])
        voxels     = np.vstack((xs, ys, zs)).T

        for ax in range(3):
            off = restrictions[ax].start
            if off is None:
                off = 0
            voxels[:, ax] += off + self.offsets[ax]

        verts, idxs = glroutines.voxelGrid(voxels, xax, yax, 1, 1)
        verts = verts.ravel('C')

        gl.glVertexPointer(3, gl.GL_FLOAT, 0, verts)
        gl.glDrawElements(gl.GL_LINES, len(idxs), gl.GL_UNSIGNED_INT, idxs)


class VoxelSelection(AnnotationObject):
    """A ``VoxelSelection`` is an :class:`AnnotationObject` which draws
    selected voxels from a :class:`.selection.Selection` instance.  A
    :class:`.SelectionTexture` is used to draw the selected voxels.
    """


    def __init__(self,
                 annot,
                 selection,
                 opts,
                 offsets=None,
                 *args,
                 **kwargs):
        """Create a ``VoxelSelection`` annotation.

        :arg annot:     The :class:`Annotations` object that owns this
                        ``VoxelSelection``.

        :arg selection: A :class:`.selection.Selection` instance which defines
                        the voxels to be highlighted.

        :arg opts:      A :class:`.NiftiOpts` instance which is used
                        for its voxel-to-display transformation matrices.

        :arg offsets:   If ``None`` (the default), the ``selection`` must have
                        the same shape as the image data being
                        annotated. Alternately, you may set ``offsets`` to a
                        sequence of three values, which are used as offsets
                        for the xyz voxel values. This is to allow for a
                        sub-space of the full image space to be annotated.

        All other arguments are passed through to the
        :meth:`AnnotationObject.__init__` method.
        """

        AnnotationObject.__init__(self, annot, *args, **kwargs)

        if offsets is None:
            offsets = [0, 0, 0]

        self.__selection = selection
        self.__opts      = opts
        self.__offsets   = offsets

        texName = '{}_{}'.format(type(self).__name__, id(selection))

        ndims = texdata.numTextureDims(selection.shape)

        if ndims == 2: ttype = textures.SelectionTexture2D
        else:          ttype = textures.SelectionTexture3D

        self.__texture = glresources.get(
            texName,
            ttype,
            texName,
            selection)


    def destroy(self):
        """Must be called when this ``VoxelSelection`` is no longer needed.
        Destroys the :class:`.SelectionTexture`.
        """
        glresources.delete(self.__texture.name)
        self.__texture = None
        self.__opts    = None


    @property
    def texture(self):
        """Return the :class:`.SelectionTexture` used by this
        ``VoxelSelection``.
        """
        return self.__texture


    def draw2D(self, zpos, axes):
        """Draws this ``VoxelSelection``."""

        xax, yax     = axes[:2]
        opts         = self.__opts
        texture      = self.__texture
        shape        = self.__selection.getSelection().shape
        displayToVox = opts.getTransform('display', 'voxel')
        voxToDisplay = opts.getTransform('voxel',   'display')
        voxToTex     = opts.getTransform('voxel',   'texture')
        voxToTex     = affine.concat(texture.texCoordXform(shape), voxToTex)
        verts, voxs  = glroutines.slice2D(shape,
                                          xax,
                                          yax,
                                          zpos,
                                          voxToDisplay,
                                          displayToVox)

        texs  = affine.transform(voxs, voxToTex)[:, :texture.ndim]
        verts = np.array(verts, dtype=np.float32).ravel('C')
        texs  = np.array(texs,  dtype=np.float32).ravel('C')

        texture.bindTexture(gl.GL_TEXTURE0)
        gl.glClientActiveTexture(gl.GL_TEXTURE0)
        gl.glTexEnvf(gl.GL_TEXTURE_ENV, gl.GL_TEXTURE_ENV_MODE, gl.GL_MODULATE)

        with glroutines.enabled((texture.target,
                                 gl.GL_TEXTURE_COORD_ARRAY,
                                 gl.GL_VERTEX_ARRAY)):
            gl.glVertexPointer(  3,            gl.GL_FLOAT, 0, verts)
            gl.glTexCoordPointer(texture.ndim, gl.GL_FLOAT, 0, texs)
            gl.glDrawArrays(     gl.GL_TRIANGLES, 0, 6)

        texture.unbindTexture()


class Text(AnnotationObject):
    """A ``Text`` is an :class:`AnnotationObject` which draws a string of
    text on the display.
    """


    def __init__(self,
                 annot,
                 text,
                 xpos,
                 ypos,
                 fontSize=10,
                 xoff=None,
                 yoff=None,
                 halign=None,
                 valign=None,
                 bgColour=None,
                 angle=None,
                 *args,
                 **kwargs):
        """Create a ``Text`` annotation.

        :arg annot:    The :class:`Annotations` object that owns this
                       ``Text``.

        :arg text:     The text to draw.

        :arg xpos:     Position along the horizontal axis as a proportion
                       between 0  (left) and 1 (right).

        :arg xpos:     Position along the vertial axis as a proportion
                       between 0 (bottom) and 1 (top).

        :arg xoff:     Fixed horizontal offset in pixels

        :arg yoff:     Fixed vertical offset in pixels

        :arg fontSize: Font size.

        :arg halign:   Horizontal alignemnt - ``'left'``, ``'centre'``, or
                       ``right``.

        :arg valign:   Vertical alignemnt - ``'bottom'``, ``'centre'``, or
                       ``top``.

        :arg bcColour: If not ``None``, a border will be drawn around the
                       text.

        :arg angle:    Angle, in degrees, by which to rotate the text.
                       NOT IMPLEMENTED YET
        """

        AnnotationObject.__init__(self, annot, *args, **kwargs)

        # We need to know the text size in pixels
        # in order to correctly align/offset the
        # text on the display. But we don't want
        # to have to calculate the size on every
        # draw. Therefore, updates to the text and
        # font size attributes are protected,
        # because they affect the final pixel text
        # size. When they are changed, we clear the
        # __textSize attribute to indicate that the
        # text size needs to be re-calculated.
        self.__text     = text
        self.__fontSize = fontSize
        self.__textSize = None

        self.xpos       = xpos
        self.ypos       = ypos
        self.xoff       = xoff
        self.yoff       = yoff
        self.bgColour   = bgColour
        self.halign     = halign
        self.valign     = valign
        self.angle      = angle


    @property
    def text(self):
        """Returns the current text value."""
        return self.__text


    @text.setter
    def text(self, value):
        """Update the text."""
        self.__text     = value
        self.__textSize = None


    @property
    def fontSize(self):
        """Returns the current font size."""
        return self.__fontSize


    @fontSize.setter
    def fontSize(self, value):
        """Update the font size."""
        self.__fontSize = value
        self.__textSize = None


    def draw2D(self, zpos, axes):
        """Draws this ``Text`` annotation. """

        canvasSize = self.annot.canvas.GetSize()
        pos        = [self.xpos * canvasSize[0], self.ypos * canvasSize[1]]

        if self.__textSize is None:
            self.__textSize = glroutines.text2D(self.text,
                                                pos,
                                                self.fontSize,
                                                canvasSize,
                                                calcSize=True)

        textSize = self.__textSize

        if   self.halign == 'centre': pos[0] -= textSize[0] / 2.0
        elif self.halign == 'right':  pos[0] -= textSize[0]

        if   self.valign == 'centre': pos[1] -= textSize[1] / 2.0
        elif self.valign == 'top':    pos[1] -= textSize[1]

        if self.xoff is not None: pos[0] += self.xoff
        if self.yoff is not None: pos[1] += self.yoff

        glroutines.text2D(self.text, pos, self.fontSize, canvasSize)



#!/usr/bin/env python
#
# dataseries.py - The DataSeries class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`DataSeries` class, the base class for
classes used by :class:`.PlotPanel` views for plotting data.
"""


import logging

import numpy as np

import fsl.utils.idle  as idle
import fsl.utils.cache as cache
import fsleyes_props   as props


log = logging.getLogger(__name__)


class DataSeries(props.HasProperties):
    """A ``DataSeries`` instance encapsulates some data to be plotted by
    a :class:`PlotPanel`, with the data extracted from an overlay in the
    :class:`.OverlayList`.

    Sub-class implementations must:

      - Accept an overlay object, :class:`.OverlayList`,
        :class:`.DisplayContext`, and :class:`.PlotPanel` in their
        ``__init__`` method, and pass these through to
        :meth:`.DataSeries.__init__`.
      - Override the :meth:`getData` method
      - Override the :meth:`redrawProperties` method if necessary


    The overlay is accessible as an instance attribute, confusingly called
    ``overlay``.


    .. note:: Some ``DataSeries`` instances may not be associated with
              an overlay (e.g. series imported loaded a text file). In
              this case, the ``overlay`` attribute will be ``None``.


    Each``DataSeries`` instance is plotted as a line, with the line
    style defined by properties on the ``DataSeries`` instance,
    such as :attr:`colour`, :attr:`lineWidth` etc.
    """


    colour = props.Colour()
    """Line colour. """


    enabled = props.Boolean(default=True)
    """Draw or not draw?"""


    alpha = props.Real(minval=0.0, maxval=1.0, default=1.0, clamped=True)
    """Line transparency."""


    label = props.String()
    """Line label (used in the plot legend)."""


    lineWidth = props.Choice((0.5, 1, 2, 3, 4, 5))
    """Line width. """


    lineStyle = props.Choice(('-',
                              '--',
                              '-.',
                              ':',
                              (0, (5, 7)),
                              (0, (1, 7)),
                              (0, (4, 10, 1, 10)),
                              (0, (4, 1, 1, 1, 1, 1)),
                              (0, (4, 1, 4, 1, 1, 1))))
    """Line style. See
    https://matplotlib.org/gallery/lines_bars_and_markers/linestyles.html
    """


    def __init__(self, overlay, overlayList, displayCtx, plotPanel):
        """Create a ``DataSeries``.

        :arg overlay:     The overlay from which the data to be plotted is
                          retrieved.  May be ``None``.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg plotPanel:   The :class:`.PlotPanel` that owns this
                          ``DataSeries``.
        """

        self.__name        = '{}_{}'.format(type(self).__name__, id(self))
        self.__overlay     = overlay
        self.__overlayList = overlayList
        self.__displayCtx  = displayCtx
        self.__plotPanel   = plotPanel
        self.setData([], [])

        log.debug('{}.init ({})'.format(type(self).__name__, id(self)))


    def __del__(self):
        """Prints a log message. """
        if log:
            log.debug('{}.del ({})'.format(type(self).__name__, id(self)))


    def __hash__(self):
        """Returns a hash for this ``DataSeries`` instance."""
        return hash(id(self))


    @property
    def name(self):
        """Returns a unique name for this ``DataSeries`` instance. """
        return self.__name


    @property
    def overlay(self):
        """Returns the overlay associated with this ``DataSeries`` instance.
        """
        return self.__overlay


    @property
    def overlayList(self):
        """Returns the :class:`.OverlayList`.
        """
        return self.__overlayList


    @property
    def displayCtx(self):
        """Returns the :class:`.DisplayContext`.
        """
        return self.__displayCtx


    @property
    def plotPanel(self):
        """Returns the :class:`.PlotPanel` that owns this ``DataSeries``
        instance.
        """
        return self.__plotPanel


    def destroy(self):
        """This method must be called when this ``DataSeries`` instance is no
        longer needed. This implementation may be overridden by sub-classes
        which need to perform any clean-up operations. Sub-class
        implementations should call this implementation.
        """
        self.__overlay     = None
        self.__overlayList = None
        self.__displayCtx  = None
        self.__plotPanel   = None


    def redrawProperties(self):
        """Returns a list of all properties which, when their values change,
        should result in this ``DataSeries`` being re-plotted. This method
        may be overridden by sub-classes.
        """

        return self.getAllProperties()[0]


    def extraSeries(self):
        """Some ``DataSeries`` types have additional ``DataSeries`` associated
        with them (see e.g. the :class:`.FEATTimeSeries` class). This method
        can be overridden to return a list of these extra ``DataSeries``
        instances. The default implementation returns an empty list.
        """
        return []


    def setData(self, xdata, ydata):
        """Set the data to be plotted. This method is irrelevant if a
        ``DataSeries`` sub-class has overridden :meth:`getData`.
        """
        self.__xdata = xdata
        self.__ydata = ydata


    def getData(self):
        """This method should be overridden by sub-classes. It must return
        the data to be plotted, as a tuple of the form:

            ``(xdata, ydata)``

        where ``xdata`` and ``ydata`` are sequences containing the x/y data
        to be plotted.

        The default implementation returns the data that has been set via the
        :meth:`setData` method.
        """
        return self.__xdata, self.__ydata


class VoxelDataSeries(DataSeries):
    """The ``VoxelDataSeries`` class is a :class:`DataSeries` class which
    provides some functionality useful to data series that represent data
    from a voxel in an :class:`.Image` overlay.

    It contains a built-in cache which is used to prevent repeated access
    to data from the same voxel.
    """


    def __init__(self, *args, **kwargs):
        """Create a ``VoxelDataSeries``. All arguments are passed through to
        the :class:`DataSeries` constructor.
        """

        DataSeries.__init__(self, *args, **kwargs)

        # We use a cache to store data for the
        # most recently accessed voxels. This is
        # done to improve performance on big
        # images (which may be compressed and
        # on disk).
        #
        # TODO You need to invalidate the cache
        #      when the image data changes.
        self.__cache = cache.Cache(maxsize=1000)


    def makeLabel(self):
        """Returns a string representation of this ``VoxelDataSeries``
        instance.
        """

        display = self.displayCtx.getDisplay(self.overlay)
        opts    = display.opts
        coords  = opts.getVoxel()

        if coords is not None:
            return '{} [{} {} {}]'.format(display.name,
                                          coords[0],
                                          coords[1],
                                          coords[2])
        else:
            return '{} [out of bounds]'.format(display.name)


    def getData(self):
        """Returns the data at the current voxel location. """

        xdata = None
        ydata = self.dataAtCurrentVoxel()

        if ydata is not None:
            xdata = np.arange(len(ydata))

        return xdata, ydata


    # The PlotPanel uses a new thread to access
    # data every time the displaycontext location
    # changes. So we mark this method as mutually
    # exclusive to prevent multiple
    # near-simultaneous accesses to the same voxel
    # location. The first time that a voxel location
    # is accessed, its data is cached. So when
    # subsequent (blocked) accesses execute, they
    # will hit the cache instead of hitting the disk
    # (which is a good thing).
    @idle.mutex
    def dataAtCurrentVoxel(self):
        """Returns the data for the current voxel of the overlay. The current
        voxel is dictated by the :attr:`.DisplayContext.location` property.
        This method is intended to be used by the :meth:`DataSeries.getData`
        method of sub-classes.

        It may also be overridden by sub-classes, but this implementation
        should be called to take advantage of the voxel data cache.

        :returns: A ``numpy`` array containing the data at the current
                  voxel, or ``None`` if the current location is out of bounds
                  of the image.
        """

        opts  = self.displayCtx.getOpts(self.overlay)
        vdim  = opts.volumeDim
        voxel = opts.getVoxel()

        if voxel is None:
            return None

        x, y, z = voxel

        data = self.__cache.get((x, y, z, vdim), None)

        if data is None:
            data = self.overlay[opts.index(voxel, atVolume=False)]
            self.__cache.put((x, y, z, vdim), data)

        return data



#!/usr/bin/env python
#
# orthoedittoolbar.py - The OrthoEditToolBar
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OrthoEditToolBar`, a
:class:`.ControlToolBar` which displays controls for editing :class:`.Image`
instances in an :class:`.OrthoPanel`.
"""


import logging

import wx

import fsleyes_props                 as props
import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.toolbar               as fsltoolbar
import fsleyes.icons                 as fslicons
import fsleyes.tooltips              as fsltooltips
import fsleyes.strings               as strings

import fsleyes.controls.displayspacewarning  as dswarning
from   fsleyes.profiles.orthoeditprofile import OrthoEditProfile


log = logging.getLogger(__name__)


class OrthoEditToolBar(ctrlpanel.ControlToolBar):
    """The ``OrthoEditToolBar`` is a :class:`.ControlToolBar` which displays
    controls for editing :class:`.Image` instances in an :class:`.OrthoPanel`.

    An ``OrthoEditToolBar`` looks something like this:


    .. image:: images/orthoedittoolbar.png
       :scale: 50%
       :align: center


    The ``OrthoEditToolBar`` exposes properties and actions which are defined
    on the :class:`.OrthoEditProfile` class, and allows the user to:

     - Change the :class:`.OrthoPanel` profile  between ``view`` and ``edit``
       mode (see the :attr:`.ViewPanel.profile` property). When in ``view``
       mode, all of the other controls are hidden.

     - Undo/redo changes to the selection and to :class:`.Image` instances.

     - Clear and fill the current selection.

     - Switch between a 2D and 3D selection cursor.

     - Change the selection cursor size.

     - Create a new mask/ROI :class:`.Image` from the current selection.

     - Switch between regular *select* mode, and *select by intensity* mode,
       and adjust the select by intensity mode settings.


    All of the controls shown on an ``OrthoEditToolBar`` instance are defined
    in the :attr:`_TOOLBAR_SPECS` dictionary.
    """


    selint = props.Boolean(default=False)
    """This property allows the user to change the :class:`.OrthoEditProfile`
    between ``sel`` mode, and ``selint`` mode.
    """


    def __init__(self, parent, overlayList, displayCtx, frame, ortho):
        """Create an ``OrthoEditToolBar``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        :arg ortho:       The :class:`.OrthoPanel` instance.
        """
        ctrlpanel.ControlToolBar.__init__(self,
                                          parent,
                                          overlayList,
                                          displayCtx,
                                          frame,
                                          height=24,
                                          kbFocus=True)

        self.__orthoPanel = ortho
        self.__dsWarning = dswarning.DisplaySpaceWarning(
            self,
            self.overlayList,
            self.displayCtx,
            self.frame,
            strings.messages[self, 'dsWarning'],
            'not like overlay',
            'overlay')

        ortho.addListener('profile', self.name, self.__profileChanged)

        self.__profileChanged()


    def destroy(self):
        """Must be called when this ``OrthoEditToolBar`` is no longer
        needed. Removes property listeners, and calls the
        :meth:`.ControlToolBar.destroy` method.
        """
        self.__orthoPanel.removeListener('profile', self.name)
        self.__dsWarning.destroy()

        self.__orthoPanel = None
        self.__dsWarning  = None

        ctrlpanel.ControlToolBar.destroy(self)


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``OrthoEditToolBar`` is only intended to be added to
        :class:`.OrthoPanel` views.
        """
        from fsleyes.views.orthopanel import OrthoPanel
        return [OrthoPanel]


    def __profileChanged(self, *a):
        """Called when the :attr:`.ViewPanel.profile` property of the
        :class:`.OrthoPanel` changes. Shows/hides edit controls accordingly.
        """

        self.ClearTools(destroy=True, postevent=False)

        ortho      = self.__orthoPanel
        profile    = ortho.profile
        profileObj = ortho.getCurrentProfile()

        if profile != 'edit':
            self.__dsWarning.Show(False)
            return

        allTools   = []
        allWidgets = []

        for specGroup in _TOOLBAR_SPECS:

            if specGroup == 'div':
                allTools.append(fsltoolbar.ToolBarDivider(self,
                                                          height=24,
                                                          orient=wx.VERTICAL))
                continue

            groupWidgets = []
            isGroup      = isinstance(specGroup, list)

            if isGroup:
                parent = wx.Panel(self)

            else:
                parent    = self
                specGroup = [specGroup]

            for spec in specGroup:

                widget = props.buildGUI(parent, profileObj, spec)

                if not isGroup and spec.label is not None:
                    widget = self.MakeLabelledTool(widget, spec.label)

                allWidgets  .append(widget)
                groupWidgets.append(widget)

            # Assuming here that all
            # widgets have labels
            if isGroup:

                sizer = wx.FlexGridSizer(2, 2, 0, 0)
                parent.SetSizer(sizer)

                labels = [s.label for s in specGroup]
                labels = [wx.StaticText(parent, label=l) for l in labels]

                for w, l in zip(groupWidgets, labels):
                    sizer.Add(l, flag=wx.ALIGN_CENTRE_VERTICAL | wx.ALIGN_RIGHT)
                    sizer.Add(w, flag=wx.ALIGN_CENTRE_VERTICAL | wx.ALIGN_RIGHT)

                allTools.append(parent)

            else:
                allTools.append(groupWidgets[0])

        self.SetTools(   [self.__dsWarning] + allTools)
        self.setNavOrder(allWidgets)


_LABELS = {

    'selectionCursorColour'  : strings.properties[OrthoEditProfile,
                                                  'selectionCursorColour'],
    'selectionOverlayColour' : strings.properties[OrthoEditProfile,
                                                  'selectionOverlayColour'],
    'selectionSize'          : strings.properties[OrthoEditProfile,
                                                  'selectionSize'],
    'intensityThres'         : strings.properties[OrthoEditProfile,
                                                  'intensityThres'],
    'searchRadius'           : strings.properties[OrthoEditProfile,
                                                  'searchRadius'],
    'fillValue'              : strings.properties[OrthoEditProfile,
                                                  'fillValue'],
    'targetImage'            : strings.properties[OrthoEditProfile,
                                                  'targetImage'],
}
"""This dictionary contains labels for some :class:`OrthoEditToolBar`
controls. It is referenced in the :attr:`_TOOLBAR_SPECS` dictionary.
"""


_ICONS = {


    'drawMode'  : [fslicons.findImageFile('drawModeHighlight24'),
                   fslicons.findImageFile('drawMode24'),
                   fslicons.findImageFile('selectModeHighlight24'),
                   fslicons.findImageFile('selectMode24')],

    'mode' : {
        'nav'    : [
            fslicons.findImageFile('addHighlight24'),
            fslicons.findImageFile('add24'),
        ],
        'sel'    : [
            fslicons.findImageFile('pencilHighlight24'),
            fslicons.findImageFile('pencil24'),
        ],
        'desel'  : [
            fslicons.findImageFile('eraserHighlight24'),
            fslicons.findImageFile('eraser24'),
        ],
        'fill' : [
            fslicons.findImageFile('bucketHighlight24'),
            fslicons.findImageFile('bucket24'),
        ],
        'selint' : [
            fslicons.findImageFile('selectByIntensityHighlight24'),
            fslicons.findImageFile('selectByIntensity24'),
        ],
    },
    'selectionIs3D' : [
        fslicons.findImageFile('selection3DHighlight24'),
        fslicons.findImageFile('selection3D24'),
        fslicons.findImageFile('selection2DHighlight24'),
        fslicons.findImageFile('selection2D24')],

    'limitToRadius' :  [
        fslicons.findImageFile('radiusHighlight24'),
        fslicons.findImageFile('radius24')],
    'localFill' :  [
        fslicons.findImageFile('localsearchHighlight24'),
        fslicons.findImageFile('localsearch24')],
    'selint' :  [
        fslicons.findImageFile('selectByIntensityHighlight24'),
        fslicons.findImageFile('selectByIntensity24')],
}
"""This dictionary contains icons for some :class:`OrthoEditToolBar`
controls. It is referenced in the :attr:`_TOOLBAR_SPECS` dictionary.
"""


_TOOLTIPS = {
    'drawMode'       : fsltooltips.properties['OrthoEditProfile.'
                                              'drawMode'],

    'mode'           : fsltooltips.properties['OrthoEditProfile.mode'],
    'selectionIs3D'  : fsltooltips.properties['OrthoEditProfile.'
                                              'selectionIs3D'],

    'limitToRadius'  : fsltooltips.properties['OrthoEditProfile.'
                                              'limitToRadius'],
    'searchRadius'   : fsltooltips.properties['OrthoEditProfile.'
                                              'searchRadius'],
    'localFill'      : fsltooltips.properties['OrthoEditProfile.localFill'],
    'selectionSize'  : fsltooltips.properties['OrthoEditProfile.'
                                              'selectionSize'],
    'fillValue'      : fsltooltips.properties['OrthoEditProfile.fillValue'],
    'intensityThres' : fsltooltips.properties['OrthoEditProfile.'
                                              'intensityThres'],

    'targetImage'    : fsltooltips.properties['OrthoEditProfile.targetImage'],
}
"""This dictionary contains tooltips for some :class:`OrthoEditToolBar`
controls. It is referenced in the :attr:`_TOOLBAR_SPECS` dictionary.
"""


def _targetImageName(image):
    if image is None: return 'None'
    else:             return image.name


_TOOLBAR_SPECS  = [

    props.Widget(
        'drawMode',
        toggle=False,
        icon=_ICONS['drawMode'],
        tooltip=_TOOLTIPS['drawMode'],
        style=wx.HORIZONTAL),
    'div',
    props.Widget(
        'mode',
        icons=_ICONS['mode'],
        tooltip=_TOOLTIPS['mode'],
        fixChoices=['nav', 'sel', 'desel', 'fill', 'selint']),
    'div',
    props.Widget(
        'selectionIs3D',
        icon=_ICONS['selectionIs3D'],
        tooltip=_TOOLTIPS['selectionIs3D'],
        toggle=False,
        dependencies=['mode'],
        enabledWhen=lambda p, m: m != 'fill'),
    props.Widget(
        'limitToRadius',
        icon=_ICONS['limitToRadius'],
        tooltip=_TOOLTIPS['limitToRadius'],
        dependencies=['mode'],
        enabledWhen=lambda p, m: m == 'selint'),
    props.Widget(
        'localFill',
        icon=_ICONS['localFill'],
        tooltip=_TOOLTIPS['localFill'],
        dependencies=['mode'],
        enabledWhen=lambda p, m: m == 'selint'),
    [props.Widget(
        'selectionSize',
        spin=True,
        slider=False,
        showLimits=False,
        spinWidth=3,
        label=_LABELS['selectionSize'],
        tooltip=_TOOLTIPS['selectionSize']),
     props.Widget(
         'fillValue',
         label=_LABELS['fillValue'],
         tooltip=_TOOLTIPS['fillValue'],
         slider=False,
         spinWidth=3,
         increment=1)],

    [props.Widget(
        'intensityThres',
        slider=True,
        spin=True,
        showLimits=False,
        spinWidth=6,
        label=_LABELS['intensityThres'],
        tooltip=_TOOLTIPS['intensityThres'],
        dependencies=['mode'],
        enabledWhen=lambda p, m: m == 'selint'),
     props.Widget(
         'searchRadius',
         slider=True,
         spin=True,
         showLimits=False,
         spinWidth=6,
         label=_LABELS['searchRadius'],
         tooltip=_TOOLTIPS['searchRadius'],
         dependencies=['mode', 'limitToRadius'],
         enabledWhen=lambda p, m, r: m == 'selint' and r)],

    props.Widget('targetImage',
                 label=_LABELS['targetImage'],
                 tooltip=_TOOLTIPS['targetImage'],
                 labels=_targetImageName,
                 dependencies=['drawMode'],
                 enabledWhen=lambda p, dm: not dm)
]
"""This list contains specifications for all of the tools shown in an
:class:`OrthoEditToolBar`, in the order that they are shown.

Some specs are grouped together into sub-lists - these will be laid out
vertically on the toolbar.
"""



#!/usr/bin/env python
#
# lightboxtoolbar.py - The LightBoxToolBar class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LightBoxToolBar` class, which is a
:class:`.ControlToolBar` for use with the :class:`.LightBoxPanel`.
"""


import wx

import fsleyes_props                 as props

import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.toolbar               as fsltoolbar
import fsleyes.actions               as actions
import fsleyes.icons                 as fslicons
import fsleyes.tooltips              as fsltooltips
import fsleyes.strings               as strings


class LightBoxToolBar(ctrlpanel.ControlToolBar):
    """The ``LightBoxToolBar`` is a :class:`.ControlToolBar` for use with the
    :class:`.LightBoxPanel`. A ``LightBoxToolBar`` looks something like this:


    .. image:: images/lightboxtoolbar.png
       :scale: 50%
       :align: center


    The ``LightBoxToolBar`` allows the user to control important parts of the
    :class:`.LightBoxPanel` display, and also to display a
    :class:`.CanvasSettingsPanel`, which allows control over all aspects of a
    ``LightBoxPanel``.
    """


    def __init__(self, parent, overlayList, displayCtx, frame, lb):
        """Create a ``LightBoxToolBar``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        :arg lb:          The :class:`.LightBoxPanel` instance.
        """

        ctrlpanel.ControlToolBar.__init__(self,
                                          parent,
                                          overlayList,
                                          displayCtx,
                                          frame,
                                          height=24,
                                          kbFocus=True)

        self.lightBoxPanel = lb

        lbOpts = lb.sceneOpts

        coronalIcon          = 'coronalSlice24'
        coronalHighlightIcon = 'coronalSliceHighlight24'

        icons = {
            'screenshot'                : fslicons.findImageFile('camera24'),
            'movieMode'                 : [
                fslicons.findImageFile('movieHighlight24'),
                fslicons.findImageFile('movie24')],
            'toggleCanvasSettingsPanel' : [
                fslicons.findImageFile('spannerHighlight24'),
                fslicons.findImageFile('spanner24')],

            'zax' : {
                0 : [fslicons.findImageFile('sagittalSliceHighlight24'),
                     fslicons.findImageFile('sagittalSlice24')],
                1 : [fslicons.findImageFile(coronalHighlightIcon),
                     fslicons.findImageFile(coronalIcon)],
                2 : [fslicons.findImageFile('axialSliceHighlight24'),
                     fslicons.findImageFile('axialSlice24')],
            }
        }

        tooltips = {

            'screenshot'   : fsltooltips.actions[   lb,      'screenshot'],
            'movieMode'    : fsltooltips.properties[lb,      'movieMode'],
            'zax'          : fsltooltips.properties[lbOpts,  'zax'],
            'sliceSpacing' : fsltooltips.properties[lbOpts,  'sliceSpacing'],
            'zrange'       : fsltooltips.properties[lbOpts,  'zrange'],
            'zoom'         : fsltooltips.properties[lbOpts,  'zoom'],
            'toggleCanvasSettingsPanel' : fsltooltips.actions[
                lb, 'toggleCanvasSettingsPanel'],
        }

        specs = {

            'toggleCanvasSettingsPanel' : actions.ToggleActionButton(
                'toggleCanvasSettingsPanel',
                actionKwargs={'floatPane' : True},
                icon=icons['toggleCanvasSettingsPanel'],
                tooltip=tooltips['toggleCanvasSettingsPanel']),

            'screenshot' : actions.ActionButton(
                'screenshot',
                icon=icons['screenshot'],
                tooltip=tooltips['screenshot']),

            'movieMode'    : props.Widget(
                'movieMode',
                icon=icons['movieMode'],
                tooltip=tooltips['movieMode']),

            'zax'          : props.Widget(
                'zax',
                icons=icons['zax'],
                tooltip=tooltips['zax']),

            'sliceSpacing' : props.Widget(
                'sliceSpacing',
                spin=False,
                showLimits=False,
                tooltip=tooltips['sliceSpacing']),

            'zrange'       : props.Widget(
                'zrange',
                spin=False,
                showLimits=False,
                tooltip=tooltips['zrange'],
                labels=[strings.choices[lbOpts, 'zrange', 'min'],
                        strings.choices[lbOpts, 'zrange', 'max']]),

            'zoom'         : props.Widget(
                'zoom',
                spin=False,
                showLimits=False,
                tooltip=tooltips['zoom']),
        }

        # Slice spacing and zoom go on a single panel
        panel = wx.Panel(self)
        sizer = wx.FlexGridSizer(2, 2, 0, 0)
        panel.SetSizer(sizer)

        more         = props.buildGUI(self,
                                      lb,
                                      specs['toggleCanvasSettingsPanel'])
        screenshot   = props.buildGUI(self,  lb,         specs['screenshot'])
        movieMode    = props.buildGUI(self,  lb,         specs['movieMode'])
        zax          = props.buildGUI(self,  lbOpts,     specs['zax'])
        zrange       = props.buildGUI(self,  lbOpts,     specs['zrange'])
        zoom         = props.buildGUI(panel, lbOpts,     specs['zoom'])
        spacing      = props.buildGUI(panel, lbOpts,     specs['sliceSpacing'])
        zoomLabel    = wx.StaticText(panel)
        spacingLabel = wx.StaticText(panel)

        zoomLabel   .SetLabel(strings.properties[lbOpts, 'zoom'])
        spacingLabel.SetLabel(strings.properties[lbOpts, 'sliceSpacing'])

        sizer.Add(zoomLabel)
        sizer.Add(zoom,    flag=wx.EXPAND)
        sizer.Add(spacingLabel)
        sizer.Add(spacing, flag=wx.EXPAND)

        tools = [more,
                 screenshot,
                 fsltoolbar.ToolBarDivider(self,
                                           height=24,
                                           orient=wx.VERTICAL),
                 zax,
                 fsltoolbar.ToolBarDivider(self,
                                           height=24,
                                           orient=wx.VERTICAL),
                 movieMode,
                 zrange,
                 panel]
        nav   = [more, screenshot, zax, movieMode, zrange, zoom, spacing]

        self.SetTools(tools)
        self.setNavOrder(nav)


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``LightBoxToolBar`` is only intended to be added to
        :class:`.LightBoxPanel` views.
        """
        from fsleyes.views.lightboxpanel import LightBoxPanel
        return [LightBoxPanel]
